from __future__ import annotations
import numpy as np
from . import util
import os

from typing import TYPE_CHECKING, Tuple
if TYPE_CHECKING:
    from typing import List, Optional as Opt, Dict, Any
    sample_t = Tuple[np.ndarray, int]


class SndInfo:
    """
    A structure to hold information about a soundfile

    Attributes:
        samplerate: the samplerate of the soundfile
        nframes: the number of frames
        channels: the number of channels per frame
        encoding: a string describing the encoding of the data.
        fileformat: the format of the soundfile ("wav", "aif", "flac", "mp3")
            This is usually the same as the extension of the soundfile read
        metadata: any metadata set in the file. Metadata, if present, is
            presented in the form of a dict[str, str], where keys are restricted
            to a subset of possible values: 'comment', 'title', 'artist',
            'album', 'tracknumber'
    """

    def __init__(self, samplerate: int, nframes: int, channels: int,
                 encoding: str, fileformat: str, metadata: Dict[str, str] = None,
                 extrainfo: Dict[str, Any] = None
                 ) -> None:
        self.samplerate: int = samplerate
        self.nframes: int = nframes
        self.channels: int = channels
        self.encoding: str = encoding
        self.fileformat: str = fileformat
        self.metadata = metadata
        self.extrainfo = extrainfo

    @property
    def duration(self) -> float:
        """ The duration of the soundfile, in seconds """
        return self.nframes / self.samplerate

    def __repr__(self):
        s = f"""samplerate : {self.samplerate}
nframes    : {self.nframes}
channels   : {self.channels}
encoding   : {self.encoding}
fileformat : {self.fileformat}
duration   : {self.duration:.3f}"""
        if self.metadata:
            s += f"\nmetadata   : {self.metadata}"
        return s


class SndWriter:
    """
    Class returned by :meth:`sndwrite_chunked` to write samples

    Args:
        backend: the Backend instance which created this SndWriter
        sr: the samplerate
        outfile: the file to wite
        encoding: the encoding of the file (*pcmXX*, *floatXX*, where *XX* represents
            the bits per sample)
        fileformat: the fileformat, only needed if the format indicated by the
            extension in outfile should be overridden
        metadata: a dict ``{str: str}``.


    **Metadata Possible Keys**:

    * title
    * comment
    * artist
    * album
    * tracknumber
    * software

    Example
    =======

    >>> from sndfileio import *
    >>> writer = sndwrite_chunked("out.flac", 44100)
    # writer is a SndWriter
    >>> for buf in sndread_chunked("in.flac"):
    ...     # do some processing, like changing the gain
    ...     buf *= 0.5
    ...     writer.write(buf)
    """

    def __init__(self, sr: int, outfile: str, encoding: str,
                 fileformat: str = None, bitrate=0,
                 backend=None,
                 metadata: Dict[str, str] = None) -> None:
        if metadata:
            for key in metadata:
                if key not in util.metadata_possible_keys:
                    raise KeyError(f"Metadata key {key} unknown. Possible keys: "
                                   f"{util.metadata_possible_keys}")
        self.sr: int = sr
        self.outfile: str = outfile
        self.encoding: str = encoding
        self.metadata: Opt[Dict[str, str]] = metadata
        self.bitrate = bitrate
        self.fileformat = fileformat or util.fileformat_from_ext(os.path.splitext(outfile)[1])
        self._backend = backend
        self._file = None

    def __call__(self, frames: np.ndarray) -> None:
        """
        Write the given sample data.

        The first array will determine the number of channels to write

        Args:
            frames (np.ndarray): the samples to write
        """
        return self.write(frames)

    def write(self, frames: np.ndarray) -> None:
        """
        Write the given sample data.

        The first array will determine the number of channels to write

        Args:
            frames (np.ndarray): the samples to write
        """
        pass

    def close(self) -> None:
        """
        Explicitely close this file
        """
        if self._file is not None:
            self._file.close()
        self._file = None

    def __enter__(self) -> SndWriter:
        return self

    def __exit__(self) -> None:
        self.close()

    @property
    def filetypes(self) -> List[str]:
        return self._backend.filetypes_write



from .datastructs import SndInfo
import pydub
import time
import numpy as np
from typing import Tuple


_cache = {}


def _audiosegment(path:str, timeout=10) -> pydub.AudioSegment:
    """
    Args:
        timeout: the time (in seconds) this audiosegment is cached
    """
    if path in _cache:
        audiosegment, t = _cache[path]
        if time.time() - t < timeout:
            return audiosegment
    audiosegment = pydub.AudioSegment.from_mp3(path)
    _cache[path] = (audiosegment, time.time())
    return audiosegment


def mp3info(path:str) -> SndInfo:
    f = _audiosegment(path)
    return SndInfo(samplerate=f.frame_rate, nframes=f.frame_count(), 
                   channels=f.channels, encoding='pcm16', fileformat='mp3')


def mp3read(path:str) -> Tuple[np.ndarray, int]:
    """
    Returns a tuple (samples, samplerate) where samples 
    is a numpy array (float, between -1 and 1)
    """
    f = _audiosegment(path)
    samples = f.get_array_of_samples()
    samplesnp = np.frombuffer(samples, np.int16, len(samples)) / (2**15)
    if f.channels > 1:
        samplesnp.shape = (int(f.frame_count()), f.channels)
    return samplesnp, f.frame_rate



"""
Helper functions for doing simple dsp filtering
"""

from __future__ import annotations
import numpy as np
import warnings
from typing import Tuple, Sequence as Seq, Union
from .util import apply_multichannel


def lowpass_cheby(samples:np.ndarray, freq:float, sr:int, maxorder=12) -> np.ndarray:
    """
    Cheby2-Lowpass Filter

    Filter data by passing data only below a certain frequency.
    The main purpose of this cheby2 filter is downsampling.
    
    This method will iteratively design a filter, whose pass
    band frequency is determined dynamically, such that the
    values above the stop band frequency are lower than -96dB.

    Args:
        samples: Data to filter, type numpy.ndarray.
        freq : The frequency above which signals are attenuated with 95 dB
        sr: Sampling rate in Hz.
        maxorder: Maximal order of the designed cheby2 filter

    Returns:
        the filtered array
    """
    if freq > sr*0.5:
        raise ValueError("Can't filter freq. above nyquist")
    b, a, freq_passband = lowpass_cheby2_coeffs(freq, sr, maxorder)
    from scipy import signal
    return signal.lfilter(b, a, samples)


# noinspection PyTupleAssignmentBalance
def lowpass_cheby2_coeffs(freq:float, sr:int, maxorder=12
                          ) -> Tuple[np.ndarray, np.ndarray, float]:
    """
    Args:
        freq: The frequency above which signals are attenuated with 95 dB
        sr: Sampling rate in Hz.
        maxorder: Maximal order of the designed cheby2 filter

    Returns:
         a tuple (b coeffs, a coeffs, freq_passband)
    """
    from scipy import signal
    nyquist = sr * 0.5
    # rp - maximum ripple of passband, rs - attenuation of stopband
    rp, rs, order = 1, 96, 1e99
    ws = freq / nyquist  # stop band frequency
    wp = ws              # pass band frequency
    # raise for some bad scenarios
    if ws > 1:
        ws = 1.0
        msg = "Selected corner frequency is above Nyquist. " + \
              "Setting Nyquist as high corner."
        warnings.warn(msg)
    while True:
        if order <= maxorder:
            break
        wp = wp * 0.99
        order, wn = signal.cheb2ord(wp, ws, rp, rs, analog=False)
    b, a = signal.cheby2(order, rs, wn, btype='low', analog=False, output='ba')
    return b, a, wp*nyquist


def filter_butter_coeffs(filtertype:str,
                         freq: Union[float, Tuple[float, float]],
                         sr:int,
                         order=5
                         ) -> Tuple[np.ndarray, np.ndarray]:
    """
    calculates the coefficients for a digital butterworth filter

    Args:
        filtertype: 'low', 'high', 'band'
        freq: cutoff freq. In the case of 'band': (low, high)
        sr: the sample rate of the data
        order: the order of the filter

    Returns:
         a tuple (b, a) with the corresponding coefficients

    """
    assert filtertype in ('low', 'high', 'band')
    from scipy import signal
    nyq = 0.5*sr
    if isinstance(freq, tuple):
        assert filtertype == 'band'
        low, high = freq
        low  /= nyq
        high /= nyq
        b, a = signal.butter(order, [low, high], btype='band', output='ba')
    else:
        freq = freq / nyq
        b, a = signal.butter(order, freq, btype=filtertype, output='ba')
    return b, a


def filter_butter(samples: np.ndarray, sr:int, filtertype:str, freq:float, order=5
                  ) -> np.ndarray:
    """
    Filters the samples with a digital butterworth filter

    Args:
        samples: mono samples
        filtertype: 'low', 'band', 'high'
        freq : for low or high, the cutoff freq; for band, (low, high)
        sr: the sampling-rate
        order: the order of the butterworth filter

    Returns:
         the filtered samples

    .. note::
    
        calls filter_butter_coeffs to calculate the coefficients
    """
    from scipy import signal
    assert filtertype in ('low', 'high', 'band')
    b, a = filter_butter_coeffs(filtertype, freq, sr, order=order)
    return apply_multichannel(samples, lambda data:signal.lfilter(b, a, data))
    

def filter_butter_plot_freqresponse(b:Seq[float], a:Seq[float],
                                    samplerate:int, numpoints=2000) -> None:
    """
    Plot the freq. response of the digital butterw. filter 
    defined by the coeffs. (b, a) at `samplerate`

    .. seealso:: filter_butter_coeffs

    """
    from scipy import signal
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError("matplotlib is needed for plotting")
    plt.figure(1)
    plt.clf()
    w, h = signal.freqz(b, a, worN=numpoints)
    plt.semilogx((samplerate*0.5/np.pi) * w, np.abs(h))
    plt.xlabel('Freq (Hz)')
    plt.ylabel('Gain')
    plt.grid(True)
    plt.show()
    


from __future__ import annotations
import numpy as np
from .datastructs import SndWriter
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import lameenc


def convertSamplesToInt16(y: np.ndarray) -> np.ndarray:
    """Convert floating-point numpy array of audio samples to int16."""
    if not issubclass(y.dtype.type, np.floating):
        raise ValueError("input samples not floating-point")
    return (y*np.iinfo(np.int16).max).astype(np.int16)


def _create_encoder(sr: int, numchannels: int, bitrate: int, quality: int
                    ) -> lameenc.Encoder:
    import lameenc
    encoder = lameenc.Encoder()
    encoder.silence()
    encoder.set_channels(numchannels)
    encoder.set_quality(quality)
    encoder.set_bit_rate(bitrate)
    encoder.set_in_sample_rate(sr)
    return encoder


class LameencWriter(SndWriter):
    def __init__(self, outfile: str, sr: int, bitrate: int, quality: int = 3):
        super().__init__(outfile=outfile, sr=sr, fileformat='mp3', encoding='pcm16')
        self._encoder = _create_encoder(sr=self.sr, numchannels=2, bitrate=bitrate,
                                        quality=quality)

    def write(self, frames: np.ndarray) -> None:
        if not self._file:
            numchannels = frames.shape[1] if len(frames.shape) == 2 else 1
            self._encoder.set_channels(numchannels)
            self._file = open(self.outfile, "wb")
        databytes = convertSamplesToInt16(frames).tobytes()
        mp3data = self._encoder.encode(databytes)
        self._file.write(mp3data)

    def close(self):
        if self._file is None:
            raise IOError("Can't close, since this file was never open")
        mp3data = self._encoder.flush()
        self._file.write(mp3data)
        self._file.close()
        self._file = None




from __future__ import annotations
from .datastructs import SndInfo
import numpy as np 
import miniaudio
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .datastructs import sample_t
    from typing import Iterator


def _readfragment(path: str, start: float, end: float
                  ) -> sample_t:
    info = miniaudio.mp3_get_file_info(path)
    seek_frame = int(start * info.sample_rate)
    if end == 0:
        frames_to_read = info.num_frames - seek_frame
    else:
        frames_to_read = int(end * info.sample_rate) - seek_frame
    buf = next(miniaudio.mp3_stream_file(path, frames_to_read=frames_to_read, seek_frame=seek_frame))
    samples = np.asarray(buf, dtype=float)
    samples /= 2**15
    return samples, info.sample_rate


def mp3read_chunked(path: str, chunksize: int, start=0., stop=0.
                    ) -> Iterator[np.ndarray]:
    info = miniaudio.mp3_get_file_info(path)
    sr = info.sample_rate
    seek_frame = int(start*sr)
    if stop == 0:
        frames_to_read = info.num_frames - seek_frame
    else:
        frames_to_read = min(info.num_frames, int(sr * stop)) - seek_frame
    assert frames_to_read > 0
    nchannels = info.nchannels
    for buf in miniaudio.mp3_stream_file(path, frames_to_read=chunksize, seek_frame=seek_frame):
        samples = np.asarray(buf, dtype=float)
        samples /= 2**15
        if nchannels > 1:
            samples.shape = (len(samples) // nchannels, nchannels)
        if frames_to_read < len(samples):
            samples = samples[:frames_to_read]
            yield samples
            return
        else:
            yield samples
            frames_to_read -= len(samples)


def mp3read(path: str, start=0., end=0.) -> sample_t:
    """
    Reads a mp3 files completely into an array
    """
    if start > 0 or end > 0:
        return _readfragment(path, start, end)
    decoded = miniaudio.mp3_read_file_f32(path)
    npsamples = np.frombuffer(decoded.samples, dtype='float32').astype(float)
    if decoded.nchannels > 1:
        npsamples.shape = (decoded.num_frames, decoded.nchannels)
    return npsamples, decoded.sample_rate


def mp3info(path: str) -> SndInfo:
    import tinytag
    info = miniaudio.mp3_get_file_info(path)
    m = tinytag.TinyTag.get(path)
    metadata = {}
    if m.title:
        metadata['title'] = m.title
    if m.album:
        metadata['album'] = m.album
    if m.comment:
        metadata['comment'] = m.comment
    if m.artist:
        metadata['artist'] = m.artist
    if m.track:
        metadata['tracknumber'] = m.track
    if not metadata:
        metadata = None
    return SndInfo(samplerate=info.sample_rate,
                   nframes=info.num_frames,
                   channels=info.nchannels,
                   encoding='pcm16',
                   fileformat='mp3',
                   metadata=metadata)
